@@ -9,14 +9,13 @@
 
 import asyncio
 import logging
-import traceback
 from collections import defaultdict
 from dataclasses import dataclass
 
 from typing import Callable, Dict, Optional, Awaitable, Union, NoReturn
 from asyncio import Task
 from aiokafka import AIOKafkaConsumer, ConsumerRecord
-from async_tools import acall
+from async_tools import acall, retry
 from kafka import TopicPartition
 
 from .abstract_kafka_client import AbstractKafkaClient
@@ -37,15 +36,17 @@ class Consumer(AbstractKafkaClient):
     class Config(AbstractKafkaClient.Config):
         topic_auto_offset_is_latest: bool = True
         group_id: Optional[str] = None
+        enable_auto_commit: bool = False
 
     def __init__(self, config: Config, exit_if_fetch_failed: bool = False) -> None:
         self.config = config
+        self._topic_to_before_last_committed_offset_callback: Dict[str, RecordCallback] = {}
         self._topic_to_callback: Dict[str, RecordCallback] = {}
         self._topic_to_call_before_fetch: Dict[str, Callable] = {}
         self._topic_to_do_before_fetch_task: Dict[str, asyncio.Future] = {}
         self._consumer: Optional[AIOKafkaConsumer] = None
         self._red_positions = defaultdict(int)
-        self._fetch_data_task: Optional[Task] = None
+        self._fetch_data__task: Optional[Task] = None
         self._exit_if_fetch_failed: bool = exit_if_fetch_failed
         super(Consumer, self).__init__(config)
 
@@ -66,6 +67,17 @@ class Consumer(AbstractKafkaClient):
         assert topic_name in self._topic_to_callback, f"no handler to topic: {topic_name}"
         self._topic_to_call_before_fetch[topic_name] = call_before_fetch
 
+    def register_call_before_last_committed_offset(self,  topic: str, callback: RecordCallback) -> None:
+        assert not self.is_connected(), "called after init"
+        self._topic_to_before_last_committed_offset_callback[topic] = callback
+
+    async def get_topic_partition_to_committed_offset(self, topic_name: str) -> dict[TopicPartition, int | None]:
+        topic_partition_to_committed_offset: dict[TopicPartition, int] = {}
+        for id_ in self._consumer.partitions_for_topic(topic_name):
+            topic_partition = TopicPartition(topic_name, id_)
+            topic_partition_to_committed_offset[topic_partition] = await self._consumer.committed(topic_partition)
+        return topic_partition_to_committed_offset
+
     async def get_topic_partition_to_max_offset(self, topic_name: str) -> Dict[TopicPartition, int]:
         topic_partitions_ids = self._consumer.partitions_for_topic(topic_name)
         topics = []
@@ -81,16 +93,37 @@ class Consumer(AbstractKafkaClient):
         self._consumer = await self._create_consumer()
         await self._consumer.start()
         await self._await_before_fetch()
+        if self._topic_to_before_last_committed_offset_callback and self.config.group_id:
+            await self._fetch_data_before_last_committed_offset()
+            logger.info("consumer finished doing callbacks before last committed offset")
         logger.info(f'consumer start listen for {self._topic_to_callback.keys()}')
-        self._fetch_data_task = asyncio.create_task(self.__fetch_data())
-        self._fetch_data_task.add_done_callback(self._fetch_finished_callback)
+        self._fetch_data__task = asyncio.create_task(self.__fetch_data())
+        self._fetch_data__task.add_done_callback(self._fetch_finished_callback)
 
-    def _fetch_finished_callback(self, fetch_task: asyncio.Task):
-        logger.warning(f"kafka fetch if finished: task: {fetch_task}")
-        # TODO fix; not working properly
-        # if fetch_task.exception() and self._exit_if_fetch_failed:
-        #     logger.critical("closing app coz kafka fetch is finished")
-        #     raise asyncio.CancelledError()
+    async def _fetch_data_before_last_committed_offset(self) -> None:
+        logger.info("consumer starts doing callbacks before last committed offset")
+        tasks = []
+        for topic, callback in self._topic_to_before_last_committed_offset_callback.items():
+            for partition, committed_offset in (await self.get_topic_partition_to_committed_offset(topic)).items():
+                if committed_offset is not None:
+                    await self._consumer.seek_to_beginning(partition)
+                    tasks.append(
+                        asyncio.create_task(
+                            self._handle_records(
+                                partition=partition, records_quantity=committed_offset, callback=callback
+                            )
+                        )
+                    )
+        await asyncio.gather(*tasks)
+
+    async def _handle_records(self, partition: TopicPartition, records_quantity: int, callback: RecordCallback) -> None:
+        for _ in range(records_quantity):
+            consumer_record = await self._consumer.getone(partition)
+            await acall(callback(consumer_record))
+
+    @staticmethod
+    def _fetch_finished_callback(fetch_task: asyncio.Task):
+        logger.warning(f"kafka fetch is finished: task: {fetch_task}")
 
     async def _create_consumer(self) -> AIOKafkaConsumer:
         auto_offset_reset = "latest" if self.config.topic_auto_offset_is_latest else "earliest"
@@ -104,15 +137,11 @@ class Consumer(AbstractKafkaClient):
         for topic, work in self._topic_to_call_before_fetch.items():
             await acall(work)
 
+    @retry
     async def __fetch_data(self) -> NoReturn:
         async for record in self._consumer:
-            try:
-                callback = self._topic_to_callback[record.topic]
-                await acall(callback(record))
-            except Exception as e:
-                logger.warning(traceback.format_exc())
-                # raise e
-            # TODO consider add option to skip business  logic exception and commit anyway
+            callback = self._topic_to_callback[record.topic]
+            await acall(callback(record))
 
     async def commit(self, record: ConsumerRecord) -> None:
         if self.config.group_id is None:
